
program PCC; (* Pascal Code Cleaner *)

uses
  Classes,
  SysUtils,
{$IFDEF USE_FLRE}
  FLRE, (* https://github.com/BeRo1985/flre *)
{$ELSE}
  RegExpr,
{$ENDIF}
  IOUtils;

function SubStr(const AStr: string; AStart: integer; AEnd: integer = MAXINT): string;
begin
  result := Copy(AStr, AStart, AEnd - AStart + 1);
end;

type
  TPascalCodeCleaner = class
    private
      FText: string;
      FData: TStringList;
    public
      constructor Create(const AFileName: string);
      destructor Destroy; override;
      procedure Extract(var ACount: integer; const AType: string; const AStart, AEnd: integer);
{$IFDEF USE_FLRE}
      (*
      function ReplaceFunc(const AInput: PFLRERawByteChar; const ACaptures: TFLRECaptures): TFLRERawByteString;
      *)
      function ReplaceFunc(const AInput: PChar; const AInputLength: Int64; const ACaptures: TFLRECaptures): AnsiString;
{$ELSE}
      function ReplaceFunc(ARegExpr: TRegExpr): string;
{$ENDIF}
      procedure Restore(const AType: string);
      procedure ExtractAll;
      procedure SaveText(const AFileName: string);
      procedure SaveData(const AFileName: string);
      procedure RestoreAll;
      procedure Remove(const AType: string);
  end;

constructor TPascalCodeCleaner.Create(const AFileName: string);
begin
  inherited Create;
  FData := TStringList.Create;
  FText := TFile.ReadAllText(AFileName);
end;

destructor TPascalCodeCleaner.Destroy;
begin
  FData.Free;
  inherited Destroy;
end;

procedure TPascalCodeCleaner.Extract(var ACount: integer; const AType: string; const AStart, AEnd: integer);
begin
  Inc(ACount);
  FData.Append(Format('_%s_%d_=%s', [AType, ACount, SubStr(FText, AStart, AEnd)]));
  FText := Format('%s_%s_%d_%s', [SubStr(FText, 1, AStart - 1), AType, ACount, SubStr(FText, AEnd + 1)]);
end;

{$IFDEF USE_FLRE}
(*
function TPascalCodeCleaner.ReplaceFunc(const AInput: PFLRERawByteChar; const ACaptures: TFLRECaptures): TFLRERawByteString;
*)
function TPascalCodeCleaner.ReplaceFunc(const AInput: PChar; const AInputLength: Int64; const ACaptures: TFLRECaptures): AnsiString;
begin
  with ACaptures[0] do result := FData.Values[FLREPtrCopy(AInput, Start, Length)];
end;
{$ELSE}
function TPascalCodeCleaner.ReplaceFunc(ARegExpr: TRegExpr): string;
begin
  with ARegExpr do result := FData.Values[Match[0]];
end;
{$ENDIF}

procedure TPascalCodeCleaner.Restore(const AType: string);
{$IFDEF USE_FLRE}
var
  LExpr: TFLRE;
begin
  LExpr := TFLRE.Create(Format('_%s_\d+_', [AType]), []);
  FText := LExpr.ReplaceCallback(FText, ReplaceFunc);
{$ELSE}
var
  LExpr: TRegExpr;
begin
  LExpr := TRegExpr.Create(Format('_%s_\d+_', [AType]));
  FText := LExpr.Replace(FText, ReplaceFunc);
{$ENDIF}
  LExpr.Free;
end;

procedure TPascalCodeCleaner.ExtractAll();
var
  c: char;
  l: integer;
  d: boolean;
  nc,
  nl,
  nd,
  i,
  iStart,
  iEnd: integer;
begin
  c := #0;
  l := 0;
  d := FALSE;
  
  nc := 0;
  nl := 0;
  nd := 0;
  
  i := 1;
  iStart := 0;
  iEnd := 0;
  
  while i <= Length(FText) do
  begin
    if c <> #0 then
    begin
      if ((c = '{') and (FText[i] = '}'))
      or ((c = '(') and (FText[i] = ')') and (FText[i - 1] = '*'))
      or ((c = '/') and (i < Length(FText)) and (FText[i + 1] in [#10, #13])) then
      begin
        iEnd := i;
        Extract(nc, 'COMMENT', iStart, iEnd);
        c := #0; // => comment = FALSE
        i := iStart - 1;
      end;
    end else

      if l <> 0 then
      begin
        if FText[i] = '''' then
        begin
          Inc(l);
          if (l mod 2 = 0) and (i < Length(FText)) and (FText[i + 1] <> '''') then
          begin
            iEnd := i;
            Extract(nl, 'LITERAL', iStart, iEnd);
            l := 0; // => literal = FALSE
            i := iStart - 1;
          end;
        end;
      end else

        if d then
        begin
          if FText[i] = '}' then
          begin
            iEnd := i;
            Extract(nd, 'DIRECTIVE', iStart, iEnd);
            d := FALSE; // => directive = FALSE
            i := iStart - 1;
          end;
        end else

          if ((FText[i] = '{') and (i < Length(FText)) and (FText[i + 1] <> '$'))
          or ((FText[i] = '(') and (i < Length(FText)) and (FText[i + 1] = '*'))
          or ((FText[i] = '/') and (i < Length(FText)) and (FText[i + 1] = '/')) then
          begin
            iStart := i;
            c := FText[i]; // => comment = TRUE
          end else

            if FText[i] = '''' then
            begin
              iStart := i;
              l := 1; // => literal = TRUE
            end else

              if ((FText[i] = '{') and (i < Length(FText)) and (FText[i + 1] = '$')) then
              begin
                iStart := i;
                d := TRUE; // => directive = TRUE
              end;

    Inc(i);
  end;
end;

procedure TPascalCodeCleaner.SaveText(const AFileName: string);
begin
  TFile.WriteAllText(AFileName, FText);
end;

procedure TPascalCodeCleaner.SaveData(const AFileName: string);
begin
  FData.SaveToFile(AFileName);
end;

procedure TPascalCodeCleaner.RestoreAll();
begin
  Restore('[A-Z]+');
end;

procedure TPascalCodeCleaner.Remove(const AType: string);
begin
{$IFDEF USE_FLRE}
  with TFLRE.Create(ansistring(Format('_%s_\d+_', [AType])), []) do
  begin
    MaximalDFAStates := 65536;
    FText := Replace(FText, '');
    Free;
  end;
{$ELSE}
  FText := ReplaceRegExpr(Format('_%s_\d+_', [AType]), FText, EmptyStr);
{$ENDIF}
end;

var
  src, dest: string;
  
begin
  if ParamCount = 2 then
  begin
    src := ParamStr(1);
    dest := ParamStr(2);
  
    if FileExists(src) then
    begin
      with TPascalCodeCleaner.Create(src) do
      try
        ExtractAll;
{$IFDEF DEBUG}
        SaveText('1.debug');
        SaveData('2.debug');
{$ENDIF}
        Remove('COMMENT');
        RestoreAll;
        SaveText(dest);
      finally
        Free;
      end;
    end;
  end;
end.
